
html {
   font-size: 14px;
}

.modal-dialog-centered-horizontally {
    display: flex;
    flex-direction: column;
    align-items: center;
    width: auto;
}

.jam-modal.modal.modal-static .modal-dialog {
    transform: none;
    transition: none;
}

.jam-modal .modal-content {
    box-shadow: 0 3px 7px rgba(0, 0, 0, 0.3);
    border: 0;
}

.dropdown-menu li {
    position: relative;
}

.dropdown-menu .dropdown-submenu {
    display: none;
    position: absolute;
    left: 100%;
    top: -7px;
}

.dropdown-menu .dropdown-submenu-left {
    right: 100%;
    left: auto;
}

.dropdown-menu > li:hover > .dropdown-submenu {
    display: block;
}

.dropdown-menu.typeahead {
    z-index: 1090;
}

.outer-table > tbody > tr {
    border-width: 0 0;
}

.outer-table > tfoot th {
    padding: 0.5rem 0.125rem;
}

.outer-table > thead > tr {
    border-width: 1px 0 1px 0;
}

.dbtable th,
.dbtable td {
    overflow-wrap: normal;
}

.inner-table tr {
    border-width: 1px 0 0 0;
    border-color: var(--bs-table-border-color);
}

.inner-table tr:first-of-type {
    border-width: 0;
}

.inner-table td {
    border-width: 0 1px;
    border-color: var(--bs-table-border-color);
    padding: 0.5rem 0.5rem;
}

.dbtable tr th:first-child,
.dbtable tr td:first-child {
    border-left-style: none;
}

.dbtable tr th:last-child,
.dbtable tr td:last-child,
.dbtable tr td:nth-last-child(2) {
    border-right-style: none;
}


.outer-table.table-bordered {
    border-style: solid;
    border-width: 0 1px 1px 1px;
    border-color: var(--bs-table-border-color);
}

div.th-container {
    overflow: hidden;
}

div.th-table {
    display: table;
    width: 100%;
    height: 100%;
}

div.th-text {
    display: table-cell;
    vertical-align: middle;
}

.dbtable td,
.dbtable th {
    cursor: default;
    -webkit-user-select: none;
    -moz-user-select: none;
    -khtml-user-select: none;
    -ms-user-select: none;
}

.dbtable td.multi-select {
    width: 4%;
}

.dbtable td > div {
    padding: 0;
}

.dbtable th.multi-select div.multi-select {
    margin-top: -0.5rem;
}

.dbtable th.multi-select .sel-count,
.dbtable th.multi-select-checkbox .sel-count {
    margin-top: 0;
    margin-bottom: 0;
    font-size: 10px;
    font-style: italic;
    color: var(--bs-primary);
}

.dbtable th.multi-select .sel-count.selected-shown {
    color: var(--bs-danger);
}

.dbtable th.multi-select .mselect-btn {
    padding: 0.125rem 0.25rem;
}

.dbtable th.multi-select .select-block {
    border: 1px solid var(--bs-secondary);
}

.dbtable th.multi-select .select-menu-btn {
    padding: 0;
    border-width: 0 0 0 1px;
}

.dbtable th.multi-select .multi-select-checkbox {
    margin: 0.375rem 0 0 0;
}

.card.jam-form {
    outline: none;
}

.tab-pane > .card.jam-form {
    border-top-width: 0;
    border-radius: 0 0 var(--bs-card-border-radius) var(--bs-card-border-radius);
}


.card.jam-form form.well   {
    border: var(--bs-border-width) var(--bs-border-style) var(--bs-border-color);
    border-radius: var(--bs-card-border-radius)
}


.card.jam-form .tab-pane form   {
    border-width: 0 var(--bs-border-width) var(--bs-border-width) var(--bs-border-width);
    border-style: var(--bs-border-style);
    border-color: var(--bs-border-color);
    border-radius: 0 0 var(--bs-card-border-radius) var(--bs-card-border-radius)
}


.tabs-div .nav-link {
    color: var(--bs-nav-tabs-link-active-color)
}

.card .form-header {
    display: flex;
    flex-shrink: 0;
    flex-wrap: nowrap;
    align-items: center;
    justify-content: flex-start;
    padding: var(--bs-card-cap-padding-y) var(--bs-card-cap-padding-x);
    margin-bottom: 0;
    color: var(--bs-card-cap-color);
    background-color: var(--bs-card-cap-bg);
    border-bottom: var(--bs-card-border-width) solid var(--bs-card-border-color);
    outline: none;
}

.card .form-header > * {
    margin: calc(var(--bs-card-cap-padding-x) * 0.3);
}

.tab-pane > .card .form-header:first-child {
  border-radius: var(--bs-card-inner-border-radius) var(--bs-card-inner-border-radius) 0 0;
}

.card .form-body {
    flex: 1 1 auto;
    padding: 0.25rem 0.25rem;
    padding: var(--bs-card-spacer-y) var(--bs-card-spacer-x);
    color: var(--bs-card-color);
    outline: none;
}

.card.form-header {
    display: flex;
    flex-direction: row;
    align-items: center;
    justify-content: space-between;
}

.form-header-close-btn,
.form-header-print-btn {
    align-self: flex-start
}

.form-header-filters {
    flex-grow: 5;
    flex-shrink: 1;
    display: flex;
    align-items: center;
}

.form-header-filters .filters-text {
    margin-left: 1rem;
}

.form-header-filters-btn {
    position: relative;
}

.form-header-filters .filtered_sign {
    position: absolute;
    top: 10%;
    right: 0.25rem;
}

h5.form-title {
    margin: 0;
    padding: 0;
}

/*
.card.no-border,
.card.no-border .form-header,
.card.no-border .form-footer {
    border-width: 0;
}
*/

/*
.card.no-border .form-body {
    padding: 0;
    border: 0;
}
*/

.card .form-footer {
    display: flex;
    flex-shrink: 0;
    flex-wrap: wrap;
    align-items: center;
    justify-content: flex-end;
    text-align: right !important;
    padding: var(--bs-card-cap-padding-y) var(--bs-card-cap-padding-x);
    color: var(--bs-card-cap-color);
    background-color: var(--bs-card-cap-bg);
    border-top: var(--bs-card-border-width) solid var(--bs-card-border-color);
}

.card .form-footer.message {
    justify-content: space-around;
}

.card .form-footer > * {
    margin: calc(var(--bs-card-cap-padding-x) * 0.3);
}

.card .form-footer #report-btn {
    padding: 0;
}

.modal-backdrop.show {
    background-color: var(--bs-white);
    opacity: 0.2;
}


.jam-modal .card .form-footer {
    background-color: var(--bs-modal-footer-bg);
}

.jam-modal .card .form-header,
 jam-modal .card .form-body {
    background-color: rgba(0, 0, 0, 0);
}

.selection-box {
    position:    absolute;
    background: transparent;
    border:     1px dotted #999;
}

.table tbody tr.row-selected > td {
    color: var(--bs-table-active-color);
    background-color: var(--bs-table-active-bg);
}

.table tbody tr.table-focused > td {
    backdrop-filter: contrast(0.9);
}

.table tbody td.field-selected {
    backdrop-filter: contrast(0.85);
}

.table tbody tr.table-focused > td.field-selected {
    backdrop-filter: contrast(0.65);
}

.table.striped tr:nth-child(odd):not(.selected) > td {
    background-color: var(--bs-table-striped-bg);
    color: var(--bs-table-striped-color)
}

.close-form-btn,
i.close-tab-btn {
    cursor: pointer;
}
.close-form-btn {
    margin-top: 0.5rem;
}

i.close-tab-btn:hover {
    color: #000;
}

td .inline-editor {
    position: absolute;
    width: calc(100% - 4px);
    height: calc(100% - 4px);
    top: 0;
    left: 0;
    margin: 2px 0 0 2px;
    border-radius: 0;
}

th .title-search-btn {
    position: absolute;
    padding: 0.25rem 0.5rem 0.25rem 0.5rem;
    top: 50%;
    transform: translate(0, -50%);
    left: 0;
    margin: 0 0 0 0.25rem;
    opacity: 0.3;
}

th .title-search-btn:hover {
    opacity: 0.6;
}

th .title-search-input-form {
    position: absolute;
    width: calc(100% - 2px);
    left: 0;
    right: 0;
    top: 50%;
    transform: translate(0, -50%);
    margin: auto;
    white-space: nowrap;
}

th .title-search-input-form input {
    padding: 0 0.375rem 0 0.375rem;
    margin: -0.25rem 0 0 0;
    border-radius: 0;
    transition: none;
}

th .title-search-input-form .title-search-input-label {
    padding: 0;
    margin: 0 0 1px 0;
    opacity: 0.7;
    font-size: 12px;
    display: block;
    margin-top: -0.25rem;
    color: var(--bs-primary);
    text-align: center;
    overflow: hidden;
}

td .hint-btn {
    position: absolute;
    padding: 0.375rem 0.5rem 0.275rem 0.5rem;
    top: 50%;
    transform: translate(0, -50%);
    right: 0;
    margin: 0 0.125rem 0 0;
    border: 0;
    background-color: rgba(0, 0, 0, 0.05);
    color: var(--bs-dark);
    font-size: 8px;
}

form.input-form {
    padding: 1rem 1rem 0.5rem 1rem;
    margin-bottom: 0.5rem;
}

.form-label.required:after {
    content: " *";
    color: var(--bs-danger);
}

.dropdown-menu.typeahead li.active {
    color: var(--bs-dropdown-link-hover-color);
    background-color: var(--bs-dropdown-link-hover-bg);
}
.datepicker td,
.datepicker th {
    width: 40px;
    height: 35px;
}

.alert-absolute {
    position: absolute;
    z-index: 3000;
    display: block;
    min-width: 300px;
}

.accordion-body p {
    margin: 0.75rem 0;
}

.accordion-body p:first-child {
    margin-top: 0
}

.accordion-body p:last-child {
    margin-bottom: 0
}

.consistent-height .tab-content {
    display: flex;
}

.consistent-height .tab-content > .tab-pane {
    display: block;
    visibility: hidden;
    margin-inline-end: -100%;
    width: 100%;
}

.consistent-height .tab-content > .active {
    visibility: visible;
}

.checkbox-centered-label {
    margin-bottom: 0.925rem;
}

.dbtree {
    padding: 2px;
    outline: none;
}

.dbtree li div {
    white-space: nowrap;
}

.dbtree li div {
    cursor: default;
    margin: 1px;
}

.dbtree li.selected > div span.tree-text {
    background-color: #e5e5e5;
}

.dbtree li.selected-focused > div span.tree-text {
    background-color: rgba(0, 0, 0, 0.1);
}

.dbtree li div span.tree-text {
    display: inline-block;
    cursor: default;
    padding: 3px 8px;
}

.dbtree ul:first-child {
    margin-left: 0px;
}

#navbar-toggler-btn {
    transition: none;
}

.paginate-scroll {
    opacity: 0.3;
    filter: alpha(opacity=30);
}

.error-modal-border {
    border: 1px solid red;
}

/* fix for new zebra datepicker plugin */
.first-btn {
    height: calc(1.5em + 0.75rem + 2px) !important;
    flex-shrink: 0;
}

/* new zebra datepicker modal form fix */
.modal-body .Zebra_DatePicker_Icon {
    top: 8px    !important;
}